@@ -23,13 +23,15 @@
 
 
 import re
+import time
 
-from gi.repository import Gtk, GLib, GtkSource, Gdk
+from gi.repository import Gtk, GLib, Gdk
 
+from inputremapper.gui.gettext import _
 from inputremapper.gui.editor.autocompletion import Autocompletion
-from inputremapper.system_mapping import system_mapping
-from inputremapper.gui.custom_mapping import custom_mapping
-from inputremapper.key import Key
+from inputremapper.configs.system_mapping import system_mapping
+from inputremapper.gui.active_preset import active_preset
+from inputremapper.event_combination import EventCombination
 from inputremapper.logger import logger
 from inputremapper.gui.reader import reader
 from inputremapper.gui.utils import CTX_KEYCODE, CTX_WARNING, CTX_ERROR
@@ -46,7 +48,7 @@ class SelectionLabel(Gtk.ListBoxRow):
 
     def __init__(self):
         super().__init__()
-        self.key = None
+        self.combination = None
         self.symbol = ""
 
         label = Gtk.Label()
@@ -62,21 +64,21 @@ class SelectionLabel(Gtk.ListBoxRow):
 
         self.show_all()
 
-    def set_key(self, key):
-        """Set the key this button represents
+    def set_combination(self, combination: EventCombination):
+        """Set the combination this button represents
 
         Parameters
         ----------
-        key : Key
+        combination : EventCombination
         """
-        self.key = key
-        if key:
-            self.label.set_label(key.beautify())
+        self.combination = combination
+        if combination:
+            self.label.set_label(combination.beautify())
         else:
-            self.label.set_label("new entry")
+            self.label.set_label(_("new entry"))
 
-    def get_key(self):
-        return self.key
+    def get_combination(self) -> EventCombination:
+        return self.combination
 
     def set_label(self, label):
         return self.label.set_label(label)
@@ -85,14 +87,14 @@ class SelectionLabel(Gtk.ListBoxRow):
         return self.label.get_label()
 
     def __str__(self):
-        return f"SelectionLabel({str(self.key)})"
+        return f"SelectionLabel({str(self.combination)})"
 
     def __repr__(self):
         return self.__str__()
 
 
 def ensure_everything_saved(func):
-    """Make sure the editor has written its changes to custom_mapping and save."""
+    """Make sure the editor has written its changes to active_preset and save."""
 
     def wrapped(self, *args, **kwargs):
         if self.user_interface.preset_name:
@@ -103,7 +105,10 @@ def ensure_everything_saved(func):
     return wrapped
 
 
-SET_KEY_FIRST = "Set the key first"
+SET_KEY_FIRST = _("Set the key first")
+
+RECORD_ALL = float("inf")
+RECORD_NONE = 0
 
 
 class Editor:
@@ -119,8 +124,11 @@ class Editor:
         self._setup_recording_toggle()
 
         self.window = self.get("window")
-        self.timeout = GLib.timeout_add(100, self.check_add_new_key)
-        self.active_selection_label = None
+        self.timeouts = [
+            GLib.timeout_add(100, self.check_add_new_key),
+            GLib.timeout_add(1000, self.update_toggle_opacity),
+        ]
+        self.active_selection_label: SelectionLabel = None
 
         selection_label_listbox = self.get("selection_label_listbox")
         selection_label_listbox.connect("row-selected", self.on_mapping_selected)
@@ -130,14 +138,7 @@ class Editor:
         # keys were not pressed yet
         self._input_has_arrived = False
 
-        toggle = self.get_recording_toggle()
-        toggle.connect("focus-out-event", self._reset_keycode_consumption)
-        toggle.connect("focus-out-event", lambda *_: toggle.set_active(False))
-        toggle.connect("toggled", self._on_recording_toggle_toggle)
-        # Don't leave the input when using arrow keys or tab. wait for the
-        # window to consume the keycode from the reader. I.e. a tab input should
-        # be recorded, instead of causing the recording to stop.
-        toggle.connect("key-press-event", lambda *args: Gdk.EVENT_STOP)
+        self.record_events_until = RECORD_NONE
 
         text_input = self.get_text_input()
         text_input.connect("focus-out-event", self.on_text_input_unfocus)
@@ -148,6 +149,21 @@ class Editor:
         target_selector = self.get_target_selector()
         target_selector.connect("changed", self._on_target_input_changed)
 
+    def __del__(self):
+        for timeout in self.timeouts:
+            GLib.source_remove(timeout)
+            self.timeouts = []
+
+    def _on_toggle_clicked(self, toggle, event=None):
+        if toggle.get_active():
+            self._show_press_key()
+        else:
+            self._show_change_key()
+
+    @ensure_everything_saved
+    def _on_toggle_unfocus(self, toggle, event=None):
+        toggle.set_active(False)
+
     @ensure_everything_saved
     def on_text_input_unfocus(self, *_):
         """When unfocusing the text it saves.
@@ -171,7 +187,7 @@ class Editor:
         One could debounce saving on text-change to avoid those logs, but that just
         sounds like a huge source of race conditions and is also hard to test.
         """
-        pass
+        pass  # the decorator will be triggered
 
     @ensure_everything_saved
     def _on_target_input_changed(self, *_):
@@ -185,18 +201,20 @@ class Editor:
         Otherwise the inputs will be read and then saved into the next preset.
         """
         if self.active_selection_label:
-            self.set_key(None)
+            self.set_combination(None)
 
-        self.set_symbol_input_text("")
-        self.disable_symbol_input()
+        self.disable_symbol_input(clear=True)
         self.set_target_selection("keyboard")  # sane default
         self.disable_target_selector()
         self._reset_keycode_consumption()
 
+        self.clear_mapping_list()
+
+    def clear_mapping_list(self):
+        """Clear the labels from the mapping selection and add an empty one."""
         selection_label_listbox = self.get("selection_label_listbox")
         selection_label_listbox.forall(selection_label_listbox.remove)
         self.add_empty()
-
         selection_label_listbox.select_row(selection_label_listbox.get_children()[0])
 
     def _setup_target_selector(self):
@@ -214,31 +232,25 @@ class Editor:
 
     def _setup_recording_toggle(self):
         """Prepare the toggle button for recording key inputs."""
-        toggle = self.get("key_recording_toggle")
-        toggle.connect(
-            "focus-out-event",
-            self._show_change_key,
-        )
-        toggle.connect(
-            "focus-in-event",
-            self._show_press_key,
-        )
-        toggle.connect(
-            "clicked",
-            lambda _: (
-                self._show_press_key()
-                if toggle.get_active()
-                else self._show_change_key()
-            ),
-        )
+        toggle = self.get_recording_toggle()
+        toggle.connect("focus-out-event", self._show_change_key)
+        toggle.connect("focus-in-event", self._show_press_key)
+        toggle.connect("clicked", self._on_toggle_clicked)
+        toggle.connect("focus-out-event", self._reset_keycode_consumption)
+        toggle.connect("focus-out-event", self._on_toggle_unfocus)
+        toggle.connect("toggled", self._on_recording_toggle_toggle)
+        # Don't leave the input when using arrow keys or tab. wait for the
+        # window to consume the keycode from the reader. I.e. a tab input should
+        # be recorded, instead of causing the recording to stop.
+        toggle.connect("key-press-event", lambda *args: Gdk.EVENT_STOP)
 
-    def _show_press_key(self, *_):
+    def _show_press_key(self, *args):
         """Show user friendly instructions."""
-        self.get("key_recording_toggle").set_label("Press Key")
+        self.get_recording_toggle().set_label("Press Key")
 
-    def _show_change_key(self, *_):
+    def _show_change_key(self, *args):
         """Show user friendly instructions."""
-        self.get("key_recording_toggle").set_label("Change Key")
+        self.get_recording_toggle().set_label("Change Key")
 
     def _setup_source_view(self):
         """Prepare the code editor."""
@@ -292,7 +304,7 @@ class Editor:
         selection_label_listbox = selection_label_listbox.get_children()
 
         for selection_label in selection_label_listbox:
-            if selection_label.get_key() is None:
+            if selection_label.get_combination() is None:
                 # unfinished row found
                 break
         else:
@@ -300,23 +312,29 @@ class Editor:
 
         return True
 
-    def disable_symbol_input(self):
+    def disable_symbol_input(self, clear=False):
         """Display help information and dont allow entering a symbol.
 
         Without this, maybe a user enters a symbol or writes a macro, switches
         presets accidentally before configuring the key and then it's gone. It can
         only be saved to the preset if a key is configured. This avoids that pitfall.
         """
+        logger.debug("Disabling the text input")
         text_input = self.get_text_input()
+
+        # beware that this also appeared to disable event listeners like
+        # focus-out-event:
         text_input.set_sensitive(False)
+
         text_input.set_opacity(0.5)
 
-        if self.get_symbol_input_text() == "":
+        if clear or self.get_symbol_input_text() == "":
             # don't overwrite user input
             self.set_symbol_input_text(SET_KEY_FIRST)
 
     def enable_symbol_input(self):
         """Don't display help information anymore and allow changing the symbol."""
+        logger.debug("Enabling the text input")
         text_input = self.get_text_input()
         text_input.set_sensitive(True)
         text_input.set_opacity(1)
@@ -340,7 +358,7 @@ class Editor:
 
     @ensure_everything_saved
     def on_mapping_selected(self, _=None, selection_label=None):
-        """One of the buttons in the left "key" column was clicked.
+        """One of the buttons in the left "combination" column was clicked.
 
         Load the information from that mapping entry into the editor.
         """
@@ -349,21 +367,21 @@ class Editor:
         if selection_label is None:
             return
 
-        key = selection_label.key
-        self.set_key(key)
+        combination = selection_label.combination
+        self.set_combination(combination)
 
-        if key is None:
-            self.set_symbol_input_text("")
-            self.disable_symbol_input()
+        if combination is None:
+            self.disable_symbol_input(clear=True)
             # default target should fit in most cases
             self.set_target_selection("keyboard")
-            # symbol input disabled until a key is configured
+            # symbol input disabled until a combination is configured
             self.disable_target_selector()
-            # symbol input disabled until a key is configured
+            # symbol input disabled until a combination is configured
         else:
-            if custom_mapping.get_mapping(key):
-                self.set_symbol_input_text(custom_mapping.get_mapping(key)[0])
-                self.set_target_selection(custom_mapping.get_mapping(key)[1])
+            if active_preset.get_mapping(combination):
+                self.set_symbol_input_text(active_preset.get_mapping(combination)[0])
+                self.set_target_selection(active_preset.get_mapping(combination)[1])
+
             self.enable_symbol_input()
             self.enable_target_selector()
 
@@ -373,22 +391,20 @@ class Editor:
         """Add one empty row for a single mapped key."""
         selection_label_listbox = self.get("selection_label_listbox")
         mapping_selection = SelectionLabel()
-        mapping_selection.set_label("new entry")
+        mapping_selection.set_label(_("new entry"))
         mapping_selection.show_all()
         selection_label_listbox.insert(mapping_selection, -1)
 
     @ensure_everything_saved
     def load_custom_mapping(self):
-        """Display the entries in custom_mapping."""
-        self.set_symbol_input_text("")
-
+        """Display the entries in active_preset."""
         selection_label_listbox = self.get("selection_label_listbox")
 
         selection_label_listbox.forall(selection_label_listbox.remove)
 
-        for key, _ in custom_mapping:
+        for key, _ in active_preset:
             selection_label = SelectionLabel()
-            selection_label.set_key(key)
+            selection_label.set_combination(key)
             selection_label_listbox.insert(selection_label, -1)
 
         self.check_add_new_key()
@@ -411,19 +427,24 @@ class Editor:
     def get_target_selector(self):
         return self.get("target-selector")
 
-    def set_key(self, key):
+    def set_combination(self, combination):
         """Show what the user is currently pressing in the user interface."""
-        self.active_selection_label.set_key(key)
+        self.active_selection_label.set_combination(combination)
 
-    def get_key(self):
-        """Get the Key object from the left column.
+        if combination and len(combination) > 0:
+            self.enable_symbol_input()
+        else:
+            self.disable_symbol_input()
+
+    def get_combination(self):
+        """Get the EventCombination object from the left column.
 
         Or None if no code is mapped on this row.
         """
         if self.active_selection_label is None:
             return None
 
-        return self.active_selection_label.key
+        return self.active_selection_label.combination
 
     def set_symbol_input_text(self, symbol):
         self.get("code_editor").get_buffer().set_text(symbol or "")
@@ -438,11 +459,11 @@ class Editor:
     def get_symbol_input_text(self):
         """Get the assigned symbol from the text input.
 
-        This might not be stored in custom_mapping yet, and might therefore also not
+        This might not be stored in active_preset yet, and might therefore also not
         be part of the preset json file yet.
 
         If there is no symbol, this returns None. This is important for some other
-        logic down the road in custom_mapping or something.
+        logic down the road in active_preset or something.
         """
         buffer = self.get("code_editor").get_buffer()
         symbol = buffer.get_text(buffer.get_start_iter(), buffer.get_end_iter(), True)
@@ -451,7 +472,7 @@ class Editor:
             # not configured yet
             return ""
 
-        return symbol
+        return symbol.strip()
 
     def set_target_selection(self, target):
         selector = self.get_target_selector()
@@ -464,20 +485,40 @@ class Editor:
         """Get a widget from the window"""
         return self.user_interface.builder.get_object(name)
 
-    def _on_recording_toggle_toggle(self, *_):
+    def update_toggle_opacity(self):
+        """If the key can't be mapped, grey it out.
+
+        During injection, when the device is grabbed and weird things are being
+        done, it is not possible.
+        """
+        toggle = self.get_recording_toggle()
+        if not self.user_interface.can_modify_preset():
+            toggle.set_opacity(0.4)
+        else:
+            toggle.set_opacity(1)
+
+        return True
+
+    def _on_recording_toggle_toggle(self, toggle):
         """Refresh useful usage information."""
-        if not self.get_recording_toggle().get_active():
+        if not toggle.get_active():
+            # if more events arrive from the time when the toggle was still on,
+            # use them.
+            self.record_events_until = time.time()
             return
+
+        self.record_events_until = RECORD_ALL
+
         self._reset_keycode_consumption()
         reader.clear()
-        if not self.user_interface.can_modify_mapping():
+        if not self.user_interface.can_modify_preset():
             # because the device is in grab mode by the daemon and
             # therefore the original keycode inaccessible
             logger.info("Cannot change keycodes while injecting")
             self.user_interface.show_status(
-                CTX_ERROR, 'Use "Stop Injection" to stop before editing'
+                CTX_ERROR, _('Use "Stop Injection" to stop before editing')
             )
-            self.get_recording_toggle().set_active(False)
+            toggle.set_active(False)
 
     def _on_delete_button_clicked(self, *_):
         """Destroy the row and remove it from the config."""
@@ -488,12 +529,12 @@ class Editor:
         ):
             return
 
-        key = self.get_key()
+        key = self.get_combination()
         if key is not None:
-            custom_mapping.clear(key)
+            active_preset.clear(key)
 
         # make sure there is no outdated information lying around in memory
-        self.set_key(None)
+        self.set_combination(None)
 
         self.load_custom_mapping()
 
@@ -501,7 +542,7 @@ class Editor:
         """Blocks until the user decided about an action."""
         confirm_delete = self.get("confirm-delete")
 
-        text = f"Are you sure to delete this mapping?"
+        text = _("Are you sure to delete this mapping?")
         self.get("confirm-delete-label").set_text(text)
 
         confirm_delete.show()
@@ -522,80 +563,97 @@ class Editor:
         if symbol != correct_case:
             self.get_text_input().get_buffer().set_text(correct_case)
 
-        # make sure the custom_mapping is up to date
-        key = self.get_key()
-        if correct_case is not None and key is not None and target is not None:
-            custom_mapping.change(key, target, correct_case)
+        # make sure the active_preset is up to date
+        key = self.get_combination()
+        if correct_case and key and target:
+            active_preset.change(key, target, correct_case)
 
         # save to disk if required
-        if custom_mapping.has_unsaved_changes():
+        if active_preset.has_unsaved_changes():
             self.user_interface.save_preset()
 
     def is_waiting_for_input(self):
-        """Check if the user is interacting with the ToggleButton for key recording."""
+        """Check if the user is trying to record buttons."""
         return self.get_recording_toggle().get_active()
 
-    def consume_newest_keycode(self, key):
-        """To capture events from keyboards, mice and gamepads.
+    def should_record_combination(self, combination):
+        """Check if the combination was written when the toggle was active."""
+        # At this point the toggle might already be off, because some keys that are
+        # used while the toggle was still on might cause the focus of the toggle to
+        # be lost, like multimedia keys. This causes the toggle to be disabled.
+        # Yet, this event should be mapped.
+        timestamp = max([event.timestamp() for event in combination])
+        return timestamp < self.record_events_until
 
-        Parameters
-        ----------
-        key : Key or None
-        """
+    def consume_newest_keycode(self, combination: EventCombination):
+        """To capture events from keyboards, mice and gamepads."""
         self._switch_focus_if_complete()
 
-        if key is None:
+        if combination is None:
             return
 
-        if not self.is_waiting_for_input():
+        if not self.should_record_combination(combination):
+            # the event arrived after the toggle has been deactivated
+            logger.debug("Recording toggle is not on")
             return
 
-        if not isinstance(key, Key):
-            raise TypeError("Expected new_key to be a Key object")
+        if not isinstance(combination, EventCombination):
+            raise TypeError("Expected new_key to be a EventCombination object")
 
         # keycode is already set by some other row
-        existing = custom_mapping.get_mapping(key)
+        existing = active_preset.get_mapping(combination)
         if existing is not None:
             existing = list(existing)
             existing[0] = re.sub(r"\s", "", existing[0])
-            msg = f'"{key.beautify()}" already mapped to "{tuple(existing)}"'
-            logger.info("%s %s", key, msg)
+            msg = _('"%s" already mapped to "%s"') % (
+                combination.beautify(),
+                tuple(existing),
+            )
+            logger.info("%s %s", combination, msg)
             self.user_interface.show_status(CTX_KEYCODE, msg)
-            return True
+            return
 
-        if key.is_problematic():
+        if combination.is_problematic():
             self.user_interface.show_status(
                 CTX_WARNING,
-                "ctrl, alt and shift may not combine properly",
-                "Your system might reinterpret combinations "
-                + "with those after they are injected, and by doing so "
-                + "break them.",
+                _("ctrl, alt and shift may not combine properly"),
+                _("Your system might reinterpret combinations ")
+                + _("with those after they are injected, and by doing so ")
+                + _("break them."),
             )
 
         # the newest_keycode is populated since the ui regularly polls it
         # in order to display it in the status bar.
-        previous_key = self.get_key()
+        previous_key = self.get_combination()
 
         # it might end up being a key combination, wait for more
         self._input_has_arrived = True
 
         # keycode didn't change, do nothing
-        if key == previous_key:
+        if combination == previous_key:
             logger.debug("%s didn't change", previous_key)
             return
 
-        self.set_key(key)
+        self.set_combination(combination)
 
         symbol = self.get_symbol_input_text()
         target = self.get_target_selection()
 
-        # the symbol is empty and therefore the mapping is not complete
-        if not symbol or not target:
+        if not symbol:
+            # has not been entered yet
+            logger.debug("Symbol missing")
+            return
+
+        if not target:
+            logger.debug("Target missing")
             return
 
         # else, the keycode has changed, the symbol is set, all good
-        custom_mapping.change(
-            new_key=key, target=target, symbol=symbol, previous_key=previous_key
+        active_preset.change(
+            new_combination=combination,
+            target=target,
+            symbol=symbol,
+            previous_combination=previous_key,
         )
 
     def _switch_focus_if_complete(self):
@@ -613,7 +671,8 @@ class Editor:
             return
 
         all_keys_released = reader.get_unreleased_keys() is None
-        if all_keys_released and self._input_has_arrived and self.get_key():
+        if all_keys_released and self._input_has_arrived and self.get_combination():
+            logger.debug("Recording complete")
             # A key was pressed and then released.
             # Switch to the symbol. idle_add this so that the
             # keycode event won't write into the symbol input as well.
