@@ -42,7 +42,7 @@ import re
 from evdev.ecodes import ecodes, EV_KEY, EV_REL, REL_X, REL_Y, REL_WHEEL, REL_HWHEEL
 
 from inputremapper.logger import logger
-from inputremapper.system_mapping import system_mapping
+from inputremapper.configs.system_mapping import system_mapping
 from inputremapper.ipc.shared_dict import SharedDict
 from inputremapper.utils import PRESS, PRESS_NEGATIVE
 
@@ -105,7 +105,7 @@ def _type_check(value, allowed_types, display_name=None, position=None):
     raise TypeError(f"Expected parameter to be one of {allowed_types}, but got {value}")
 
 
-def _type_check_keyname(keyname):
+def _type_check_symbol(keyname):
     """Same as _type_check, but checks if the key-name is valid."""
     if isinstance(keyname, Variable):
         # it is a variable and will be read at runtime
@@ -136,6 +136,8 @@ def _type_check_variablename(name):
 def _resolve(argument, allowed_types=None):
     """If the argument is a variable, figure out its value and cast it.
 
+    Variables are prefixed with `$` in the syntax.
+
     Use this just-in-time when you need the actual value of the variable
     during runtime.
     """
@@ -200,12 +202,6 @@ class Macro:
 
         self.running = False
 
-        # all required capabilities, without those of child macros
-        self.capabilities = {
-            EV_KEY: set(),
-            EV_REL: set(),
-        }
-
         self.child_macros = []
 
         self.keystroke_sleep_ms = None
@@ -248,20 +244,6 @@ class Macro:
         """Check if the macro is waiting for a key to be released."""
         return not self._trigger_release_event.is_set()
 
-    def get_capabilities(self):
-        """Get the merged capabilities of the macro and its children."""
-        capabilities = copy.deepcopy(self.capabilities)
-
-        for macro in self.child_macros:
-            macro_capabilities = macro.get_capabilities()
-            for ev_type in macro_capabilities:
-                if ev_type not in capabilities:
-                    capabilities[ev_type] = set()
-
-                capabilities[ev_type].update(macro_capabilities[ev_type])
-
-        return capabilities
-
     async def run(self, handler):
         """Run the macro.
 
@@ -280,8 +262,7 @@ class Macro:
         # newly arriving events are only interesting if they arrive after the
         # macro started
         self._new_event_arrived.clear()
-
-        self.keystroke_sleep_ms = self.context.mapping.get("macros.keystroke_sleep_ms")
+        self.keystroke_sleep_ms = self.context.preset.get("macros.keystroke_sleep_ms")
 
         self.running = True
         for task in self.tasks:
@@ -317,24 +298,40 @@ class Macro:
             macro.release_trigger()
 
     async def _keycode_pause(self, _=None):
-        """To add a pause between keystrokes."""
-        await asyncio.sleep(self.keystroke_sleep_ms / 1000)
+        """To add a pause between keystrokes.
 
-    def add_mouse_capabilities(self):
-        """Add all capabilities that are required to recognize the device as mouse."""
-        self.capabilities[EV_REL].add(REL_X)
-        self.capabilities[EV_REL].add(REL_Y)
-        self.capabilities[EV_REL].add(REL_WHEEL)
-        self.capabilities[EV_REL].add(REL_HWHEEL)
+        This was needed at some point because it appeared that injecting keys too
+        fast will prevent them from working. It probably depends on the environment.
+        """
+        await asyncio.sleep(self.keystroke_sleep_ms / 1000)
 
     def __repr__(self):
         return f'<Macro "{self.code}">'
 
     """Functions that prepare the macro"""
 
+    def add_key(self, symbol):
+        """Write the symbol."""
+        # This is done to figure out if the macro is broken at compile time, because
+        # if KEY_A was unknown we can show this in the gui before the injection starts.
+        _type_check_symbol(symbol)
+
+        async def task(handler):
+            # if the code is $foo, figure out the correct code now.
+            resolved_symbol = _resolve(symbol, [str])
+            code = _type_check_symbol(resolved_symbol)
+
+            resolved_code = _resolve(code, [int])
+            handler(EV_KEY, resolved_code, 1)
+            await self._keycode_pause()
+            handler(EV_KEY, resolved_code, 0)
+            await self._keycode_pause()
+
+        self.tasks.append(task)
+
     def add_hold(self, macro=None):
         """Loops the execution until key release."""
-        _type_check(macro, [Macro, str, None], "h (hold)", 1)
+        _type_check(macro, [Macro, str, None], "hold", 1)
 
         if macro is None:
             self.tasks.append(lambda _: self._trigger_release_event.wait())
@@ -343,16 +340,18 @@ class Macro:
         if not isinstance(macro, Macro):
             # if macro is a key name, hold down the key while the
             # keyboard key is physically held down
-            code = _type_check_keyname(macro)
+            symbol = macro
+            _type_check_symbol(symbol)
 
             async def task(handler):
+                resolved_symbol = _resolve(symbol, [str])
+                code = _type_check_symbol(resolved_symbol)
+
                 resolved_code = _resolve(code, [int])
-                self.capabilities[EV_KEY].add(resolved_code)
                 handler(EV_KEY, resolved_code, 1)
                 await self._trigger_release_event.wait()
                 handler(EV_KEY, resolved_code, 0)
 
-            self.capabilities[EV_KEY].add(code)
             self.tasks.append(task)
 
         if isinstance(macro, Macro):
@@ -374,31 +373,46 @@ class Macro:
         modifier : str
         macro : Macro
         """
-        _type_check(macro, [Macro], "m (modify)", 2)
-
-        modifier = str(modifier)
-        code = system_mapping.get(modifier)
-
-        if code is None:
-            raise KeyError(f'Unknown modifier "{modifier}"')
-
-        self.capabilities[EV_KEY].add(code)
+        _type_check(macro, [Macro], "modify", 2)
+        _type_check_symbol(modifier)
 
         self.child_macros.append(macro)
 
         async def task(handler):
-            resolved_code = _resolve(code, [int])
-            self.capabilities[EV_KEY].add(resolved_code)
-            await self._keycode_pause()
-            handler(EV_KEY, resolved_code, 1)
+            # TODO test var
+            resolved_modifier = _resolve(modifier, [str])
+            code = _type_check_symbol(resolved_modifier)
+
+            handler(EV_KEY, code, 1)
             await self._keycode_pause()
             await macro.run(handler)
             await self._keycode_pause()
-            handler(EV_KEY, resolved_code, 0)
+            handler(EV_KEY, code, 0)
             await self._keycode_pause()
 
         self.tasks.append(task)
 
+    def add_hold_keys(self, *symbols):
+        """Hold down multiple keys, equivalent to `a + b + c + ...`."""
+        for symbol in symbols:
+            _type_check_symbol(symbol)
+
+        async def task(handler):
+            resolved_symbols = [_resolve(symbol, [str]) for symbol in symbols]
+            codes = [_type_check_symbol(symbol) for symbol in resolved_symbols]
+
+            for code in codes:
+                handler(EV_KEY, code, 1)
+                await self._keycode_pause()
+
+            await self._trigger_release_event.wait()
+
+            for code in codes[::-1]:
+                handler(EV_KEY, code, 0)
+                await self._keycode_pause()
+
+        self.tasks.append(task)
+
     def add_repeat(self, repeats, macro):
         """Repeat actions.
 
@@ -407,8 +421,8 @@ class Macro:
         repeats : int or Macro
         macro : Macro
         """
-        repeats = _type_check(repeats, [int], "r (repeat)", 1)
-        _type_check(macro, [Macro], "r (repeat)", 2)
+        repeats = _type_check(repeats, [int], "repeat", 1)
+        _type_check(macro, [Macro], "repeat", 2)
 
         async def task(handler):
             for _ in range(_resolve(repeats, [int])):
@@ -417,55 +431,27 @@ class Macro:
         self.tasks.append(task)
         self.child_macros.append(macro)
 
-    def add_key(self, symbol):
-        """Write the symbol."""
-        _type_check_keyname(symbol)
-
-        symbol = str(symbol)
-        code = system_mapping.get(symbol)
-        self.capabilities[EV_KEY].add(code)
-
-        async def task(handler):
-            handler(EV_KEY, code, 1)
-            await self._keycode_pause()
-            handler(EV_KEY, code, 0)
-            await self._keycode_pause()
-
-        self.tasks.append(task)
-
-    def add_event(self, _type, code, value):
+    def add_event(self, type_, code, value):
         """Write any event.
 
         Parameters
         ----------
-        _type: str or int
+        type_: str or int
             examples: 2, 'EV_KEY'
-        code : int or int
+        code : str or int
             examples: 52, 'KEY_A'
         value : int
         """
-        _type = _type_check(_type, [int, str], "e (event)", 1)
-        code = _type_check(code, [int, str], "e (event)", 2)
-        value = _type_check(value, [int, str], "e (event)", 3)
+        type_ = _type_check(type_, [int, str], "event", 1)
+        code = _type_check(code, [int, str], "event", 2)
+        value = _type_check(value, [int, str], "event", 3)
 
-        if isinstance(_type, str):
-            _type = ecodes[_type.upper()]
+        if isinstance(type_, str):
+            type_ = ecodes[type_.upper()]
         if isinstance(code, str):
             code = ecodes[code.upper()]
 
-        if _type not in self.capabilities:
-            self.capabilities[_type] = set()
-
-        if _type == EV_REL:
-            # add all capabilities that are required for the display server
-            # to recognize the device as mouse
-            self.capabilities[EV_REL].add(REL_X)
-            self.capabilities[EV_REL].add(REL_Y)
-            self.capabilities[EV_REL].add(REL_WHEEL)
-
-        self.capabilities[_type].add(code)
-
-        self.tasks.append(lambda handler: handler(_type, code, value))
+        self.tasks.append(lambda handler: handler(type_, code, value))
         self.tasks.append(self._keycode_pause)
 
     def add_mouse(self, direction, speed):
@@ -480,8 +466,6 @@ class Macro:
             "right": (REL_X, 1),
         }[direction.lower()]
 
-        self.add_mouse_capabilities()
-
         async def task(handler):
             resolved_speed = value * _resolve(speed, [int])
             while self.is_holding():
@@ -502,8 +486,6 @@ class Macro:
             "right": (REL_HWHEEL, -1),
         }[direction.lower()]
 
-        self.add_mouse_capabilities()
-
         async def task(handler):
             resolved_speed = _resolve(speed, [int])
             while self.is_holding():
@@ -535,7 +517,7 @@ class Macro:
 
         self.tasks.append(task)
 
-    def add_ifeq(self, variable, value, then=None, otherwise=None):
+    def add_ifeq(self, variable, value, then=None, else_=None):
         """Old version of if_eq, kept for compatibility reasons.
 
         This can't support a comparison like ifeq("foo", $blub) with blub containing
@@ -543,7 +525,7 @@ class Macro:
         variable name.
         """
         _type_check(then, [Macro, None], "ifeq", 3)
-        _type_check(otherwise, [Macro, None], "ifeq", 4)
+        _type_check(else_, [Macro, None], "ifeq", 4)
 
         async def task(handler):
             set_value = macro_variables.get(variable)
@@ -551,20 +533,20 @@ class Macro:
             if set_value == value:
                 if then is not None:
                     await then.run(handler)
-            elif otherwise is not None:
-                await otherwise.run(handler)
+            elif else_ is not None:
+                await else_.run(handler)
 
         if isinstance(then, Macro):
             self.child_macros.append(then)
-        if isinstance(otherwise, Macro):
-            self.child_macros.append(otherwise)
+        if isinstance(else_, Macro):
+            self.child_macros.append(else_)
 
         self.tasks.append(task)
 
-    def add_if_eq(self, value_1, value_2, then=None, _else=None):
+    def add_if_eq(self, value_1, value_2, then=None, else_=None):
         """Compare two values."""
         _type_check(then, [Macro, None], "if_eq", 3)
-        _type_check(_else, [Macro, None], "if_eq", 4)
+        _type_check(else_, [Macro, None], "if_eq", 4)
 
         async def task(handler):
             resolved_value_1 = _resolve(value_1)
@@ -572,17 +554,17 @@ class Macro:
             if resolved_value_1 == resolved_value_2:
                 if then is not None:
                     await then.run(handler)
-            elif _else is not None:
-                await _else.run(handler)
+            elif else_ is not None:
+                await else_.run(handler)
 
         if isinstance(then, Macro):
             self.child_macros.append(then)
-        if isinstance(_else, Macro):
-            self.child_macros.append(_else)
+        if isinstance(else_, Macro):
+            self.child_macros.append(else_)
 
         self.tasks.append(task)
 
-    def add_if_tap(self, then=None, _else=None, timeout=300):
+    def add_if_tap(self, then=None, else_=None, timeout=300):
         """If a key was pressed quickly.
 
         macro key pressed -> if_tap starts -> key released -> then
@@ -591,13 +573,13 @@ class Macro:
         -> if_tap starts -> pressed -> released -> then
         """
         _type_check(then, [Macro, None], "if_tap", 1)
-        _type_check(_else, [Macro, None], "if_tap", 2)
+        _type_check(else_, [Macro, None], "if_tap", 2)
         timeout = _type_check(timeout, [int, float], "if_tap", 3)
 
         if isinstance(then, Macro):
             self.child_macros.append(then)
-        if isinstance(_else, Macro):
-            self.child_macros.append(_else)
+        if isinstance(else_, Macro):
+            self.child_macros.append(else_)
 
         async def wait():
             """Wait for a release, or if nothing pressed yet, a press and release."""
@@ -614,20 +596,20 @@ class Macro:
                 if then:
                     await then.run(handler)
             except asyncio.TimeoutError:
-                if _else:
-                    await _else.run(handler)
+                if else_:
+                    await else_.run(handler)
 
         self.tasks.append(task)
 
-    def add_if_single(self, then, otherwise, timeout=None):
+    def add_if_single(self, then, else_, timeout=None):
         """If a key was pressed without combining it."""
         _type_check(then, [Macro, None], "if_single", 1)
-        _type_check(otherwise, [Macro, None], "if_single", 2)
+        _type_check(else_, [Macro, None], "if_single", 2)
 
         if isinstance(then, Macro):
             self.child_macros.append(then)
-        if isinstance(otherwise, Macro):
-            self.child_macros.append(otherwise)
+        if isinstance(else_, Macro):
+            self.child_macros.append(else_)
 
         async def task(handler):
             triggering_event = (self._newest_event.type, self._newest_event.code)
@@ -663,7 +645,7 @@ class Macro:
             except asyncio.TimeoutError:
                 pass
 
-            if otherwise:
-                await otherwise.run(handler)
+            if else_:
+                await else_.run(handler)
 
         self.tasks.append(task)
